//----------------------------------------------------------------------------
//
// File:        cartridge.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1994-2000 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include "common.hpp"
#include "logger.hpp"
#include "compress.hpp"
#include "cartridge.hpp"

DBG_REGISTER ( __FILE__ );

const char *cCartridge::sm_Banner = "TI-99/4A Module - ";

void cCartridge::LoadImage ( char *filename )
{
    FUNCTION_ENTRY ( this, "cCartridge::LoadImage", true );

    FILE *file = filename ? fopen ( filename, "rb" ) : NULL;
    if ( file == NULL ) return;

    char buffer [ 80 ];
    fread ( buffer, sizeof ( buffer ), 1, file );
    if ( strncmp ( buffer, sm_Banner, strlen ( sm_Banner ))) return;
    char *ptr = &buffer [ strlen ( sm_Banner )];
    ptr [ strlen ( ptr ) - 2 ] = '\0';
    m_Title = new char [ strlen ( ptr ) + 1 ];
    strcpy ( m_Title, ptr );

    EVENT ( "Loading module: " << m_Title );

    UCHAR region = ( UCHAR ) fgetc ( file );
    while ( ! feof ( file )) {
        sMemoryRegion *memory;
        bool compressed = ( region & 0x80 ) ? true : false;
        bool dsr = ( region & 0x40 ) ? true : false;
        region &= 0x3F;
        memory = ( region < GROM_0 ) ? &CpuMemory [ region ] : &GromMemory [ region - GROM_0 ];
        if ( dsr ) {
            fread ( &memory->cruBase, 1, sizeof ( memory->cruBase ), file );
            EVENT ( "  CRU Base: " << ( void * ) memory->cruBase );
        } else {
            memory->cruBase = 0;
        }
        USHORT totalSize = ( USHORT ) (( region < GROM_0 ) ? 0x1000 : 0x2000 );
        memory->Type  = ( eMemoryType ) fgetc ( file );
        memory->noBanks = ( USHORT ) fgetc ( file );
        fread ( memory->noBytes, 1, sizeof ( memory->noBytes ), file );

        EVENT ( "  " << (( region < GROM_0 ) ? "RAM" : "GROM" ) << " @ " << ( void * ) ( region * totalSize ));

        for ( int i = 0; i < memory->noBanks; i++ ) {
            memory->Bank[i] = new UCHAR [ totalSize ];
            memset ( memory->Bank[i], 0, totalSize );
            if ( memory->Type == _ROM_ ) {
                if ( compressed ) {
                    LoadBuffer ( memory->noBytes[i], memory->Bank[i], file );
                } else {
                    fread ( memory->Bank[i], memory->noBytes[i], 1, file );
                }
            }
        }
        memory->curBank = memory->Bank[0];
        region = ( UCHAR ) fgetc ( file );
    }
    fclose ( file );
}

void cCartridge::SaveImage ( char *filename )
{
    FUNCTION_ENTRY ( this, "cCartridge::SaveImage", true );

    FILE *file = filename ? fopen ( filename, "wb" ) : NULL;
    if ( file == NULL ) return;

    char buffer [ 80 ];
    memset ( buffer, 0, sizeof ( buffer ));
    sprintf ( buffer, "%s%s\n%c", sm_Banner, m_Title, 0x1A );
    fwrite ( buffer, sizeof ( buffer ), 1, file );

    EVENT ( "Saving module: " << m_Title );

    for ( unsigned i = 0; i < SIZE ( CpuMemory ); i++ ) {
        if ( CpuMemory[i].noBanks != 0 ) {
            sMemoryRegion *memory = &CpuMemory[i];
            int cru = memory->cruBase ? 0x40 : 0x00;
            fputc ( 0x80 | cru | ROM_0 + i, file );
            if ( cru ) fwrite ( &memory->cruBase, 1, sizeof ( memory->cruBase ), file );
            fputc ( memory->Type, file );
            fputc ( memory->noBanks, file );
            fwrite ( memory->noBytes, 1, sizeof ( memory->noBytes ), file );
            if ( memory->Type == _ROM_ ) {
                for ( int j = 0; j < memory->noBanks; j++ ) {
                    SaveBuffer ( memory->noBytes[j], memory->Bank[j], file );
                }
            }
        }
    }

    for ( unsigned i = 0; i < SIZE ( GromMemory ); i++ ) {
        if ( GromMemory[i].noBanks != 0 ) {
            sMemoryRegion *memory = &GromMemory[i];
            int cru = memory->cruBase ? 0x40 : 0x00;
            fputc ( 0x80 | cru | GROM_0 + i, file );
            if ( cru ) fwrite ( &memory->cruBase, 1, sizeof ( memory->cruBase ), file );
            fputc ( memory->Type, file );
            fputc ( memory->noBanks, file );
            fwrite ( memory->noBytes, 1, sizeof ( memory->noBytes ), file );
            if ( memory->Type == _ROM_ ) {
                for ( int j = 0; j < memory->noBanks; j++ ) {
                    SaveBuffer ( memory->noBytes[j], memory->Bank[j], file );
                }
            }
        }
    }

    fclose ( file );
}

cCartridge::cCartridge ( char *filename )
{
    FUNCTION_ENTRY ( this, "cCartridge ctor", true );

    for ( unsigned i = 0; i < SIZE ( CpuMemory ); i++ ) {
        CpuMemory[i].Type = _ROM_;
        CpuMemory[i].cruBase = 0;
        CpuMemory[i].noBanks = 0;
        CpuMemory[i].curBank = NULL;
        for ( unsigned j = 0; j < 4; j++ ) {
            CpuMemory[i].noBytes[j] = 0;
            CpuMemory[i].Bank[j] = NULL;
        }
    }
    for ( unsigned i = 0; i < SIZE ( GromMemory ); i++ ) {
        GromMemory[i].Type = _ROM_;
        GromMemory[i].cruBase = 0;
        GromMemory[i].noBanks = 0;
        GromMemory[i].curBank = NULL;
        for ( unsigned j = 0; j < 4; j++ ) {
            GromMemory[i].noBytes[j] = 0;
            GromMemory [i].Bank[j] = NULL;
        }
    }
    m_Title = NULL;
    LoadImage ( filename );
}

cCartridge::~cCartridge ()
{
    FUNCTION_ENTRY ( this, "cCartridge dtor", true );

    for ( unsigned i = 0; i < SIZE ( CpuMemory ); i++ ) {
        for ( unsigned j = 0; j < CpuMemory[i].noBanks; j++ ) {
            delete [] CpuMemory[i].Bank[j];
        }
    }

    for ( unsigned i = 0; i < SIZE ( GromMemory ); i++ ) {
        for ( unsigned j = 0; j < GromMemory[i].noBanks; j++ ) {
            delete [] GromMemory[i].Bank[j];
        }
    }

    memset ( CpuMemory, 0, sizeof ( CpuMemory ));
    memset ( GromMemory, 0, sizeof ( GromMemory ));

    if ( m_Title ) delete [] m_Title;
}

bool cCartridge::IsValid () const
{
    FUNCTION_ENTRY ( this, "cCartridge::IsValid", true );

    for ( unsigned i = 0; i < SIZE ( CpuMemory ); i++ ) {
        if ( CpuMemory[i].noBanks > 0 ) return true;
    }
    for ( unsigned i = 0; i < SIZE ( GromMemory ); i++ ) {
        if ( GromMemory[i].noBanks > 0 ) return true;
    }
    return false;
}
