//----------------------------------------------------------------------------
//
// File:        tms9918a.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1994-2000 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <memory.h>
#include <stdio.h>
#include "common.hpp"
#include "logger.hpp"
#include "compress.hpp"
#include "tms9900.hpp"
#include "tms9918a.hpp"
#include "ti994a.hpp"

extern void Panic ( char * );

cTMS9918A::cTMS9918A ()
{
    FUNCTION_ENTRY ( this, "cTMS9918A ctor", true );

    m_Memory = new UCHAR [ 0x4000 ];
    memset ( m_Memory, 0, 0x4000 );

    m_ImageTable      = ( sScreenImage * ) m_Memory;
    m_ColorTable      = ( sColorTable * ) m_Memory;
    m_PatternTable    = ( sPatternDescriptor * ) m_Memory;
    m_SpriteAttrTable = ( sSpriteAttribute * ) m_Memory;
    m_SpriteDescTable = ( sSpriteDescriptor * ) m_Memory;

    m_ColorTableSize   = 0x0020;
    m_PatternTableSize = 0x0800;
    m_ImageTableSize   = 0x0300;

    m_Host = NULL;

    m_Address  = 0;
    m_Transfer = 0;
    m_Shift    = 0;

    memset ( &m_Status, 0, sizeof ( m_Status ));
    memset ( m_Register, 0, sizeof ( m_Register ));

    m_Mode = 0;
}

cTMS9918A::~cTMS9918A ()
{
    FUNCTION_ENTRY ( this, "cTMS9918A dtor", true );

    delete m_Memory;
}

void cTMS9918A::SetHost ( cTI994A *host )
{
    FUNCTION_ENTRY ( this, "cTMS9918A::SetHost", true );

    m_Host = host;
}

void cTMS9918A::Reset ()
{
    FUNCTION_ENTRY ( this, "cTMS9918A::Reset", true );

    memset ( m_Memory, 0, 0x4000 );
    memset ( &m_Status, 0, sizeof ( m_Status ));

    // Set mode to ZERO here so we don't actually do any mode switch stuff
    m_Mode    = 0xFF;
    m_Address = m_Transfer = m_Shift = 0;
    memset ( m_Register, -1, sizeof ( m_Register ));
    for ( unsigned i = 0; i < SIZE ( m_Register ); i++ ) {
        WriteRegister ( i, 0 );
    }
}

bool wrap = false;

extern int  HistoryIndex;
extern char HistoryBuffer [1024][80];

void cTMS9918A::WriteAddress ( UCHAR data )
{
    FUNCTION_ENTRY ( this, "cTMS9918A::WriteAddress", true );

    if ( m_Shift == 0 ) {
        m_Transfer &= 0xFF00;
        m_Transfer |= data;
        m_Shift = 8;
    } else {
        m_Transfer &= 0x00FF;
        m_Transfer |= ( USHORT ) ( data << 8 );
        m_Shift = 0;
        if ( m_Transfer & 0x8000 ) {
            if ( data & 0x08 ) {
                ERROR ( "Unexpected register value: " << hex << data );
            }
            WriteRegister ( data & 0x07, ( UCHAR ) m_Transfer );
        } else {
            m_Address = m_Transfer;
            wrap = false;
        }
    }
}

void cTMS9918A::WriteData ( UCHAR data )
{
    FUNCTION_ENTRY ( this, "cTMS9918A::WriteData", true );

    if (( m_Address & 0x4000 ) == 0 ) {
        WARNING ( "** Writing to VDP RAM with read address" );
    }
    if ( wrap ) {
        WARNING ( "Address wrapped around" );
        wrap = false;
    }
    if (( m_Address & 0x3FFF ) == 0x3FFF ) {
        wrap = true;
    }

    m_Memory [ m_Address++ & 0x3FFF ] = data;
}

UCHAR cTMS9918A::ReadData ()
{
    FUNCTION_ENTRY ( this, "cTMS9918A::ReadData", true );

    if ( m_Address & 0x4000 ) {
        WARNING ( "** Reading from VDP RAM with write address" );
    }
    if ( wrap ) {
        WARNING ( "Address wrapped around" );
        wrap = false;
    }
    if (( m_Address & 0x3FFF ) == 0x3FFF ) {
        wrap = true;
    }

    return m_Memory [ m_Address++ & 0x3FFF ];
}

void cTMS9918A::WriteRegister ( int reg, UCHAR value )
{
    FUNCTION_ENTRY ( this, "cTMS9918A::WriteRegister", true );

    static UCHAR mask [8] = { 0xFF, 0xFF, 0x0F, 0xFF, 0x07, 0x7F, 0x07, 0xFF };

    ASSERT (( reg >= 0 ) && ( reg < 8 ));

    // Mask off unused bits - TI-99/4A ROM sets these bits to 1s
    value &= mask [reg];
    m_Register [reg] = value;

    int offset, newMode = m_Mode;

    switch ( reg ) {
        case 0 :
            m_ColorTableSize   = ( value & VDP_MODE_3_BIT ) ? 0x1800 : 0x0020;
            m_PatternTableSize = ( value & VDP_MODE_3_BIT ) ? 0x1800 : 0x0800;
            newMode &= ~ VDP_M3;
            if ( value & VDP_MODE_3_BIT ) newMode |= VDP_M3;
            if ( newMode != m_Mode ) {
                value  = m_Register [3] & mask [3];
                offset = ( newMode & VDP_M3 ) ? ( value & 0x80 ) ? 0x2000 : 0 : value * 0x0040;
                ASSERT ( offset < 0x4000 );
                m_ColorTable = ( sColorTable * ) &m_Memory [ offset ];
                value = m_Register [4] & mask [4];
                offset = ( newMode & VDP_M3 ) ? ( value & 0x04 ) ? 0x2000 : 0 : value * 0x0800;
                ASSERT ( offset < 0x4000 );
                m_PatternTable = ( sPatternDescriptor * ) &m_Memory [ offset ];
            }
            SetMode ( newMode );
            break;
        case 1 :
            m_ImageTableSize = ( value & VDP_MODE_1_BIT ) ? 0x03C0 : 0x0300;
            newMode &= ~ ( VDP_M2 | VDP_M1 );
            if ( value & VDP_MODE_2_BIT ) newMode |= VDP_M2;
            if ( value & VDP_MODE_1_BIT ) newMode |= VDP_M1;
            SetMode ( newMode );
            break;
        case 2 :
            m_ImageTable = ( sScreenImage *) &m_Memory [ value * 0x0400 ];
            break;
        case 3 :
            offset = ( m_Mode & VDP_M3 ) ? ( value & 0x80 ) ? 0x2000 : 0 : value * 0x0040;
            m_ColorTable = ( sColorTable * ) &m_Memory [ offset ];
            break;
        case 4 :
            offset = ( m_Mode & VDP_M3 ) ? ( value & 0x04 ) ? 0x2000 : 0 : value * 0x0800;
            ASSERT ( offset < 0x4000 );
            m_PatternTable = ( sPatternDescriptor * ) &m_Memory [ offset ];
            break;
        case 5 :
            ASSERT ( value * 0x0080 < 0x4000 );
            m_SpriteAttrTable = ( sSpriteAttribute * ) &m_Memory [ value * 0x0080 ];
            break;
        case 6 :
            ASSERT ( value * 0x0800 < 0x4000 );
            m_SpriteDescTable = ( sSpriteDescriptor * ) &m_Memory [ value * 0x0800 ];
            break;
    }
}

UCHAR cTMS9918A::ReadStatus ()
{
    FUNCTION_ENTRY ( this, "cTMS9918A::ReadStatus", true );

    UCHAR retVal = * ( UCHAR * ) &m_Status;
    memset ( &m_Status, 0, sizeof ( m_Status ));
    return retVal;
}

bool cTMS9918A::SetMode ( int newMode )
{
    FUNCTION_ENTRY ( this, "cTMS9918A::SetMode", true );

    if ( newMode == m_Mode ) return false;
    m_Mode = ( UCHAR ) ( newMode & ( VDP_M3 | VDP_M2 | VDP_M1 ));
    return true;
}

void cTMS9918A::Retrace ()
{
    FUNCTION_ENTRY ( this, "cTMS9918A::Retrace", true );

    if ( InterruptsEnabled ()) {
        m_Status.InterruptFlag = 1;
        if ( m_Host ) m_Host->VerticalRetrace ();
    }
}

void cTMS9918A::LoadImage ( FILE *file )
{
    FUNCTION_ENTRY ( this, "cTMS9918A::LoadImage", true );

    fread ( &m_Address, sizeof ( m_Address ), 1, file );
    fread ( &m_Transfer, sizeof ( m_Transfer ), 1, file );
    fread ( &m_Shift, sizeof ( m_Shift ), 1, file );

    UCHAR NewRegister [8];
    fread ( &m_Status, sizeof ( m_Status ), 1, file );
    fread ( NewRegister, sizeof ( NewRegister ), 1, file );

    LoadBuffer ( 0x4000, m_Memory, file );

    m_Mode = 0xFF;	// Guarantee that SetMode will be called at least once
    for ( unsigned i = 0; i < 8; i++ ) {
        WriteRegister ( i, NewRegister [i] );
    }
}

void cTMS9918A::SaveImage ( FILE *file )
{
    FUNCTION_ENTRY ( this, "cTMS9918A::SaveImage", true );

    fwrite ( &m_Address, sizeof ( m_Address ), 1, file );
    fwrite ( &m_Transfer, sizeof ( m_Transfer ), 1, file );
    fwrite ( &m_Shift, sizeof ( m_Shift ), 1, file );

    fwrite ( &m_Status, sizeof ( m_Status ), 1, file );
    fwrite ( m_Register, sizeof ( m_Register ), 1, file );

    SaveBuffer ( 0x4000, m_Memory, file );
}
