//----------------------------------------------------------------------------
//
// File:        ti994a-sdl.cpp
// Date:        18-Apr-2000
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 2000 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include "SDL.h"
#include "SDL_thread.h"
#include "common.hpp"
#include "logger.hpp"
#include "bitmap.hpp"
#include "compress.hpp"
#include "cartridge.hpp"
#include "tms9900.hpp"
#include "tms9918a.hpp"
#include "tms9919.hpp"
#include "ti994a.hpp"
#include "ti994a-sdl.hpp"

#define K_BKSP		0x008
#define K_TAB		0x009
#define K_ENT		0x00D
#define K_ESC		0x01B
#define K_SPC		0x020
#define K_DEL		0x07F
#define K_NUMLCK	0x12C
#define K_CAPLCK	0x12D
#define K_RSHF		0x12F
#define K_LSHF		0x130
#define K_SHF		0x130
#define K_RCTR		0x131
#define K_LCTR		0x132
#define K_CTR		0x132
#define K_RALT		0x133
#define K_LALT		0x134
#define K_ALT		0x134

#define K_UP		0x111
#define K_DN		0x112
#define K_RT		0x113
#define K_LT		0x114

#define K_HOME		0x116
#define K_END		0x117
#define K_PGUP		0x118
#define K_PGDN		0x119

#define K_INS		0x100

#define K_0		0x030
#define K_1		0x031
#define K_2		0x032
#define K_3		0x033
#define K_4		0x034
#define K_5		0x035
#define K_6		0x036
#define K_7		0x037
#define K_8		0x038
#define K_9		0x039

#define K_A		0x061
#define K_B		0x062
#define K_C		0x063
#define K_D		0x064
#define K_E		0x065
#define K_F		0x066
#define K_G		0x067
#define K_H		0x068
#define K_I		0x069
#define K_J		0x06A
#define K_K		0x06B
#define K_L		0x06C
#define K_M		0x06D
#define K_N		0x06E
#define K_O		0x06F
#define K_P		0x070
#define K_Q		0x071
#define K_R		0x072
#define K_S		0x073
#define K_T		0x074
#define K_U		0x075
#define K_V		0x076
#define K_W		0x077
#define K_X		0x078
#define K_Y		0x079
#define K_Z		0x07A

#define K_QUO		0x027		// '"
#define K_COM		0x02C		// ,<
#define K_MIN		0x02D		// -_
#define K_PER		0x02E		// .>
#define K_DIV		0x02F		// /?
#define K_SEM		0x03B		// ;:
#define K_EQL		0x03D		// =+
#define K_LBR		0x05B		// [{
#define K_BAR		0x05C		// \|
#define K_RBR		0x05D		// ]}
#define K_TIL		0x060		// `~

static cSdlTI994A *pThis;

cSdlTI994A::cSdlTI994A ( cCartridge *ctg, cTMS9918A *vdp, cTMS9919 *sound ) :
    cTI994A ( ctg, vdp, sound ),
    m_StartTime ( 0 ),
    m_StopTime ( 0 )
{
    FUNCTION_ENTRY ( this, "cSdlTI994A ctor", true );

    m_Running  = false;    
    m_CapsLock = false;
    m_ColumnSelect = 0;
    memset ( m_StateTable, 0, sizeof ( m_StateTable ));

    pThis = this;
}

cSdlTI994A::~cSdlTI994A ()
{
    FUNCTION_ENTRY ( this, "cSdlTI994A dtor", true );
}

bool cSdlTI994A::isSpecial ( USHORT ch )
{
    FUNCTION_ENTRY ( NULL, "cSdlTI994A::isSpecial", true );

    switch ( ch ) {
        case K_ESC  : 			// ESC
        case K_TIL  : 			// `~
        case K_MIN  : 			// -_
        case K_BKSP : 			// Back Space
        case K_BAR  : 			// \|
        case K_LBR  : 			// [{
        case K_RBR  : 			// ]}
        case K_QUO  : 			// '"
        case K_DIV  :			// /?
        case K_DEL  :			// DEL
            return true;
        case K_UP : 			// UP
        case K_DN : 			// DOWN
        case K_LT :	 		// LEFT
        case K_RT :			// RIGHT
            return ( m_StateTable [ K_NUMLCK ] != 0 ) ? false : true;
	default :
            return false;
    }
}

void cSdlTI994A::SetSpecial ( USHORT ch )
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::SetSpecial", true );

    int shifted = m_StateTable [ K_SHF ] & 0x0F;

    bool fctn = false, shift = false;
    switch ( ch ) {
        case K_TIL : if ( shifted ) {		// ~
                        fctn = true;
                        ch = K_W;
                    }
		    break;
        case K_MIN : if ( shifted ) {		// _
                        fctn = true;
                        ch = K_U;
	            } else {			// -
                        shift = true;
                        ch = K_DIV;
		    }
                    break;
        case K_BAR : if ( shifted ) {		// |
                        fctn = true;
                        ch = K_A;
                    }
                    break;
        case K_LBR : fctn = true;		// [{
                    ch = ( USHORT ) ( shifted ? K_F : K_R );
                    break;
        case K_RBR : fctn = true;		// ]}
                    ch = ( USHORT ) ( shifted ? K_G : K_T );
                    break;
        case K_QUO : fctn = true;
                    ch = ( USHORT ) ( shifted ? K_P : K_O );
                    break;
        case K_DIV : if ( shifted ) {
                        fctn = true;
                        ch = K_I;
                    }
                    break;
        case K_UP : fctn = true;
                    ch = K_E;
                    break;
        case K_DN : fctn = true;
                    ch = K_X;
                    break;
        case K_BKSP :
        case K_LT : fctn = true;
                    ch = K_S;
                    break;
        case K_RT : fctn = true;
                    ch = K_D;
                    break;
        case K_DEL : fctn = true;
                    ch = K_1;
                    break;
    }
    if ( shifted && ! shift ) m_StateTable [ K_SHF ] = 0;
    if ( ch ) m_StateTable [ ch ] |= 0xF0;
    if ( fctn ) m_StateTable [ K_ALT ] |= 0xF0;
    if ( shift ) m_StateTable [ K_SHF ] |= 0xF0;
}

void cSdlTI994A::ClearSpecial ( USHORT ch )
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::ClearSpecial", true );

    USHORT ch1 = 0, ch2 = 0;
    switch ( ch ) {
        case K_TIL : ch1 = K_W;
		    break;
        case K_MIN : ch1 = K_U;
                    ch2 = K_DIV;
                    break;
        case K_BAR : ch1 = K_A;
                    break;
        case K_LBR : ch1 = K_F;
                    ch2 = K_R;
                    break;
        case K_RBR : ch1 = K_G;
                    ch2 = K_T;
                    break;
        case K_QUO : ch1 = K_P;
                    ch2 = K_O;
                    break;
        case K_DIV : ch1 = K_DIV;
                    ch2 = K_I;
                    break;
        case K_UP : ch1 = K_E;
                    break;
        case K_DN : ch1 = K_X;
                    break;
        case K_BKSP :
        case K_LT : ch1 = K_S;
                    break;
        case K_RT : ch1 = K_D;
                    break;
        case K_DEL : ch1 = K_1;
                    break;
    }
    m_StateTable [ K_ALT ] &= 0x0F;
    m_StateTable [ K_SHF ] &= 0x0F;
    if ( ch1 ) m_StateTable [ ch1 ] &= 0x0F;
    if ( ch2 ) m_StateTable [ ch2 ] &= 0x0F;
}

void cSdlTI994A::KeyPressed ( USHORT ch )
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::KeyPressed", true );

    if ( isSpecial ( ch )) {
        SetSpecial ( ch );
    } else {
        UCHAR val = 0x0F;
        switch ( ch ) {
            case K_LSHF : ch = K_SHF;	val = 0x0C;  break;
            case K_RSHF : ch = K_SHF;	val = 0x03;  break;
            case K_LCTR : ch = K_CTR;	val = 0x0C;  break;
            case K_RCTR : ch = K_CTR;	val = 0x03;  break;
            case K_LALT : ch = K_ALT;	val = 0x0C;  break;
            case K_RALT : ch = K_ALT;	val = 0x03;  break;
        }
        m_StateTable [ ch ] |= val;
    }
}

void cSdlTI994A::KeyReleased ( USHORT ch )
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::KeyReleased", true );

    if ( isSpecial ( ch )) {
        ClearSpecial ( ch );
    } else {
        UCHAR mask = 0xF0;
        switch ( ch ) {
            case K_LSHF : ch = K_SHF;	mask = 0xF3;  break;
            case K_RSHF : ch = K_SHF;	mask = 0xFC;  break;
            case K_LCTR : ch = K_CTR;	mask = 0xF3;  break;
            case K_RCTR : ch = K_CTR;	mask = 0xFC;  break;
            case K_LALT : ch = K_ALT;	mask = 0xF3;  break;
            case K_RALT : ch = K_ALT;	mask = 0xFC;  break;
        }
        m_StateTable [ ch ] &= mask;
    }
}

/*
    >00  0 = Internal 9901 Control   1 = Clock Control
    >01  Set by an external Interrupt
    >02  Set by TMS9918A on Vertical Retrace Interrupt
    >03  Set by Clock Interrupt for Cassette read/write routines
    >0C  Reserved - High Level
    >16  Cassette CS1 motor control On/Off
    >17  Cassette CS2 motor control On/Off
    >18  Audio Gate enable/disable
    >19  Cassette Tape Out
    >1B  Cassette Tape In
*/
 
int cSdlTI994A::ReadCRU ( unsigned short address )
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::ReadCRU", true );

    // TI Keyboard Matrix - Table of IBM scan-codes
    static USHORT Keys[8][8] = {
        { K_EQL, K_PER, K_COM,   K_M,   K_N, K_DIV, K_INS,  0 },
        { K_SPC,   K_L,   K_K,   K_J,   K_H, K_SEM,  K_LT,  0 },
        { K_ENT,   K_O,   K_I,   K_U,   K_Y,   K_P,  K_RT,  0 },
        {     0,   K_9,   K_8,   K_7,   K_6,   K_0,  K_DN,  0 },
        { K_ALT,   K_2,   K_3,   K_4,   K_5,   K_1,  K_UP,  0 },
        { K_SHF,   K_S,   K_D,   K_F,   K_G,   K_A,     0,  0 },
        { K_CTR,   K_W,   K_E,   K_R,   K_T,   K_Q,     0,  0 },
        {     0,   K_X,   K_C,   K_V,   K_B,   K_Z,     0,  0 }
    };
    int retVal = 1;
    if ( address == 2 ) {
        retVal = m_VDP_Retrace ? 1 : 0;
    } else if (( address >= 3 ) && ( address <= 10 )) {
        if (( m_CapsLock == false ) && ( address == 7 )) return m_StateTable [ K_CAPLCK ] ? 1 : 0;

        USHORT index = Keys [ address - 3 ][ m_ColumnSelect ];
        retVal = ( m_StateTable [ index ] != 0 ) ? 0 : 1;

        switch ( m_ColumnSelect ) {
            case 6:			// Joystick 1
                switch ( address ) {
                    case 9 : if ( m_StateTable [ K_HOME ] || m_StateTable [ K_END ]  ) retVal = 0;	break;
                    case 8 : if ( m_StateTable [ K_PGUP ] || m_StateTable [ K_PGDN ] ) retVal = 0;	break;
                    case 7 : if ( m_StateTable [ K_END ]  || m_StateTable [ K_PGDN ] ) retVal = 0;	break;
                    case 6 : if ( m_StateTable [ K_HOME ] || m_StateTable [ K_PGUP ] ) retVal = 0;	break;
                }
                break;
            case 7:			// Joystick 2
                switch ( address ) {
                    case 9 : if ( m_StateTable [ K_HOME ] || m_StateTable [ K_END ]  ) retVal = 0;	break;
                    case 8 : if ( m_StateTable [ K_PGUP ] || m_StateTable [ K_PGDN ] ) retVal = 0;	break;
                    case 7 : if ( m_StateTable [ K_END ]  || m_StateTable [ K_PGDN ] ) retVal = 0;	break;
                    case 6 : if ( m_StateTable [ K_HOME ] || m_StateTable [ K_PGUP ] ) retVal = 0;	break;
                }
                break;
            default :
                break;
        }
    } else {
        retVal = cTI994A::ReadCRU ( address );
    }

    return retVal;
}

void cSdlTI994A::WriteCRU ( ADDRESS address, USHORT data )
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::WriteCRU", true );

    if ( address == 2 ) {
        if ( m_VDP_Retrace ) m_VDP_Retrace--;
    } else if (( address >= 18 ) && ( address <= 20 )) {
        int shift = address - 18;
        m_ColumnSelect &= ~ ( 1 << shift );
        m_ColumnSelect |= data << shift;
    } else if ( address == 21 ) {
        m_CapsLock = ( data != 0 ) ? true : false;
    } else {
        cTI994A::WriteCRU ( address, data );
    }
}

int cSdlTI994A::_TimerHookProc ()
{
    FUNCTION_ENTRY ( NULL, "cSdlTI994A::_TimerHookProc", true );

    return pThis->TimerHookProc ();
}

int cSdlTI994A::TimerHookProc ()
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::TimerHookProc", true );

    static ULONG lastClock = 0;
    static ULONG startClock = 0;

    ULONG clockCycles    = m_CPU->GetClocks ();

    ULONG ellapsedCycles = clockCycles - startClock;
    ULONG ellapsedTime   = SDL_GetTicks () - m_StartTime;

    if ( ellapsedTime > 0 ) {
        // Limit the emulated speed to 3.3MHz
        if ( ellapsedCycles / ellapsedTime > 3.3 * 1000.0 ) {
            SDL_Delay ( 1 );
        }

        // Reset base time/clocks every 5 minutes to avoid wrap
        if ( ellapsedTime > 5 * 60 * 1000.0 ) {
            startClock  = clockCycles;
            m_StartTime = SDL_GetTicks ();
        }
    }

    // Simulate a 60Hz VDP interrupt
    if ( clockCycles - lastClock > 3300000 / 60 ) {
        lastClock = clockCycles;
        RetraceTimer ();
    }

    return 0;
}

int cSdlTI994A::_RunThreadProc ( void *ptr )
{
    FUNCTION_ENTRY ( NULL, "cSdlTI994A::_RunThreadProc", true );

    cSdlTI994A *pThis = ( cSdlTI994A * ) ptr;

    return pThis->RunThreadProc ();
}

extern "C" ULONG TimerHook;

int cSdlTI994A::RunThreadProc ()
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::RunThreadProc", true );

    TimerHook = ( ULONG ) _TimerHookProc;

    m_CPU->Run ();

    return 0;
}

void cSdlTI994A::StartThread ()
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::StartThread", true );

    if ( m_Running == true ) return;

    m_Running = true;

    m_StartTime = SDL_GetTicks () - m_StopTime;
    
    m_Thread = SDL_CreateThread ( _RunThreadProc, this );
}

void cSdlTI994A::StopThread ()
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::StopThread", true );

    if ( m_Running == false ) return;

    m_Running = false;

    m_CPU->Stop ();

    m_StopTime = SDL_GetTicks ();

    SDL_WaitThread ( m_Thread, NULL );
}

void cSdlTI994A::Reset ()
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::Reset", true );

    StopThread ();

    cTI994A::Reset ();

    StartThread ();
}

void cSdlTI994A::Run ()
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::Run", true );

    StartThread ();

    // Loop waiting for SDL_QUIT
    SDL_Event event;
    while ( SDL_WaitEvent ( &event ) >= 0 ) {
        switch ( event.type ) {
            case SDL_KEYDOWN :
                switch ( event.key.keysym.sym ) {
                    case SDLK_ESCAPE :
                        goto done;
                    case SDLK_F10 :
                        Reset ();
                        break;
                    default :
                        KeyPressed ( event.key.keysym.sym );
                        break;
		}
                break;
            case SDL_KEYUP :
                KeyReleased ( event.key.keysym.sym );
                break;
            case SDL_QUIT:
                goto done;
        }
    }

done:
    
    StopThread ();

}

void cSdlTI994A::RetraceTimer ()
{
    FUNCTION_ENTRY ( this, "cSdlTI994A::RetraceTimer", true );

    m_VDP->Retrace ();
    m_VDP->Refresh ();
}
